
def minDist(atomList1, atomList2):
    dMin = 9999999.99999
    for v1 in [ numpy.array((a.x ,a.y, a.z)) for a in atomList1 ]:
        for v2 in [ numpy.array((b.x ,b.y, b.z)) for b in atomList2 ]:
            d = numpy.linalg.norm(v1 - v2)
            if d < dMin:
                dMin = d
    return dMin


def euclidianDist(atom1, atom2):
    mtx_atom1 = numpy.array(atom1.coordinates)
    mtx_atom2 = numpy.array(atom2.coordinates)
    return numpy.linalg.norm(mtx_atom1 - mtx_atom2)


class Cell(object):
    def __init__(self, iLabel, jLabel, value):
        self.iLabel = iLabel
        self.jLabel = jLabel
        self.value = value

    def __str__(self):
        return '(' + self.iLabel + ', ' + self.jLabel + ') : ' + str(self.value)



class MeshMap(object):
    def __init__(self, s1, s2):
        pass


class ContactMap(object):
    def __init__(self, s1, s2):
        self.nb_raws = s1.residueNumber
        self.nb_columns = s2.residueNumber
        self.mtx = numpy.zeros((self.nb_raws, self.nb_columns))

        self._resArrayOne = [ x for x in s1.byres() ]
        self._resArrayTwo = [ x for x in s2.byres() ]

        self.rl = [r.id for r in self._resArrayOne]
        self.cl = [r.id for r in self._resArrayTwo]

        for i1, r1 in enumerate(self._resArrayOne):
            for i2, r2 in enumerate(self._resArrayTwo):
                self.mtx[i1, i2] = minDist(r1, r2)

    # matrix element accessor along with row/column label, mostly for inspection
    def __getitem__(self, tup):
        x, y = tup
        return Cell(self.rl[x], self.cl[y], self.mtx[x, y])

    def __str__(self):
        asString = 'Contact map ' + str(len(self.rl)) + 'x' + str(len(self.cl)) + '\n'
        asString += '\t\t' + ''.join(["%9s" % r for r in self.cl]) + '\n'
        for i, row in enumerate(self.mtx):
            asString += "%9s" % self.rl[i]
            asString += ' '.join(["%9.1f" % d for d in row])
            asString += '\n'
        return asString

    def Q(self, d=4.5):
        f = lambda x: 1 if x < d else 0
        DistToQ = numpy.vectorize(f)
        return DistToQ(self.mtx)

    def residuesInterfacialBool(self, d=4.5):
        q = self.Q(d)
        def f(row):
            if numpy.sum(row) > 0:
                return 1
            return 0

        return interfaceBoolList( [({ 'num' : r.num, 'name' : r.name, 'chain' : r.chain }, f(q[i])) for i,r in enumerate(self._resArrayOne)],
            [({ 'num' : r.num, 'name' : r.name, 'chain' : r.chain}, f(q.T[i])) for i,r in enumerate(self._resArrayTwo)] )

    def weighted_contact_number(self):
        mtx = numpy.zeros((self.nb_raws, self.nb_columns))

        for i in range(len(self._resArrayOne)):
            for j in range(len(self._resArrayTwo)):
                if i != j:
                    mtx[i, j] = 1/(self.mtx[i, j]**2)

        return mtx.sum(axis=1,dtype=float)


class ContactMap_intra(object):
    def __init__(self, struc, cutoff=5.0):
        self.nb = struc.residueNumber
        self.mtx = numpy.zeros((self.nb,self.nb))

        self._resArray = [ x for x in struc.byres() ]

        self.l = [r.id for r in self._resArray]

        self.counter_infcutoff = 0

        for i in range(len(self._resArray)):
            for j in range(i+1,len(self._resArray)):
                self.mtx[i, j] = minDist(self._resArray[i], self._resArray[j])

                if self.mtx[i, j] < cutoff:
                    self.counter_infcutoff += 1

    # matrix element accessor along with row/column label, mostly for inspection
    def __getitem__(self, tup):
        x, y = tup
        return Cell(self.l[x], self.l[y], self.mtx[x, y])

    def __str__(self):
        asString = 'Contact map intra ' + str(len(self.l)) + 'x' + str(len(self.l)) + '\n'
        asString += '\t\t' + ''.join(["%9s" % r for r in self.l]) + '\n'
        for i, row in enumerate(self.mtx):
            asString += "%9s" % self.l[i]
            asString += ' '.join(["%9.1f" % d for d in row])
            asString += '\n'
        return asString

    def weighted_contact_number(self):
        mtx = numpy.zeros((self.nb,self.nb))

        for i in range(len(self._resArray)):
            for j in range(i+1,len(self._resArray)):
                mtx[i, j] = 1/(self.mtx[i, j]**2)

        symetric_mtx = mtx + mtx.T

        return symetric_mtx.sum(axis=1,dtype=float)


class ContactMap_intra_grid(object):
    def __init__(self, struc, cutoff=5.0):
        self.struc = struc
        self._resArray = self.struc.getResID
        self._residuePairRegistry = {}
        self.cutoff = cutoff
        self._parsing()
        self._build_grid()
        self._list_next_unique_neighbors()
        self._calculate_distances()
        self._build_ContactMap()

    # matrix element accessor along with row/column label, mostly for inspection
    def __getitem__(self, tup):
        x, y = tup
        return Cell(self._resArray[x], self._resArray[y], self.mtx[x, y])

    def __str__(self):
        asString = 'Contact map intra grid ' + str(len(self._resArray)) + 'x' + str(len(self._resArray)) + '\n'
        asString += '\t\t' + ''.join(["%9s" % r for r in self._resArray]) + '\n'
        for i, row in enumerate(self.mtx):
            asString += "%9s" % self._resArray[i]
            asString += ' '.join(["%9.1f" % d for d in row])
            asString += '\n'
        return asString

    def _parsing(self):
        # ---
        # PARSING
        # ---

        self.counter_atoms = 0

        self.min_x,self.min_y,self.min_z = 9999999.99999,9999999.99999,9999999.99999
        self.max_x,self.max_y,self.max_z = -9999999.99999,-9999999.99999,-9999999.99999

        for atom in self.struc.atomRecord:
            # ---
            # GET MINIMUMS AND MAXIMUMS OF EACH COORDINATES X,Y,Z
            # ---

            self.counter_atoms += 1

            self.min_x = atom.x if atom.x < self.min_x else self.min_x
            self.min_y = atom.y if atom.y < self.min_y else self.min_y
            self.min_z = atom.z if atom.z < self.min_z else self.min_z

            self.max_x = atom.x if atom.x > self.max_x else self.max_x
            self.max_y = atom.y if atom.y > self.max_y else self.max_y
            self.max_z = atom.z if atom.z > self.max_z else self.max_z

        # ---
        # GET DIMENSIONS OF THE 3D GRID
        # ---

        self.dim_x = int(math.floor((self.max_x-self.min_x)/self.cutoff))+1
        self.dim_y = int(math.floor((self.max_y-self.min_y)/self.cutoff))+1
        self.dim_z = int(math.floor((self.max_z-self.min_z)/self.cutoff))+1

    def _coor_Cartesian2Grid(self,atom):
        # ---
        # GET AN ATOM AND RETURN ITS NORMALIZED COORDINATES
        # ---
        return [int(math.floor((atom.x-self.min_x)/self.cutoff)),int(math.floor((atom.y-self.min_y)/self.cutoff)),int(math.floor((atom.z-self.min_z)/self.cutoff))]

    def _build_grid(self):
        self.grid_3D = numpy.empty((self.dim_x,self.dim_y,self.dim_z),dtype=numpy.object_)
        self.grid_3D.fill([])
        self.grid_3D = numpy.frompyfunc(list,1,1)(self.grid_3D)

        for atom in self.struc.atomRecord:
            [i,j,k] = self._coor_Cartesian2Grid(atom)
            self.grid_3D[i,j,k].append(atom)

    def _iter_next_unique_neighbors(self, x, y, z):
        value_list = []

        for i in range(x, x+2):
            if i == self.dim_x:
                break

            for j in range(y-1, y+2):
                if j < 0:
                    continue
                if j == self.dim_y:
                    break

                for k in range(z-1, z+2):
                    if k < 0:
                        continue
                    if k == self.dim_z:
                        break

                    if ( i > x or j > y or (j > y-1 and k > z-1) ) and self.grid_3D[i,j,k] != []:
                        value_list.append(self.grid_3D[i,j,k])

        return value_list

    def _list_next_unique_neighbors(self):
        self.value_LIST = []

        for i in range(self.dim_x):
            for j in range(self.dim_y):
                for k in range(self.dim_z):

                    if self.grid_3D[i,j,k] != []:
                        self.value_LIST.append(self._iter_next_unique_neighbors(i,j,k))

    def _calculate_distances(self):
        self.dist_LIST = []
        self.counter_dist = 0

        for i in range(len(self.value_LIST)):
            current = self.value_LIST[i][0]
            neighbors_flat_LIST = [j for i2 in self.value_LIST[i][1:] for j in i2]

            for j in range(len(current)):

                for k in range(j+1,len(current)):
                    self.dist_LIST.append([[current[j],current[k]],euclidianDist(current[j],current[k])])
                    self.counter_dist += 1

                for l in range(len(neighbors_flat_LIST)):
                    self.dist_LIST.append([[current[j],neighbors_flat_LIST[l]],euclidianDist(current[j],neighbors_flat_LIST[l])])
                    self.counter_dist += 1

    def _build_ContactMap(self):
        self.mtx = numpy.empty((len(self._resArray),len(self._resArray)),dtype=float)
        self.mtx.fill(9999999.99999)

        for idx in self.dist_LIST:
            i = self._resArray.index(idx[0][0].getResID)
            j = self._resArray.index(idx[0][1].getResID)
            if idx[1] > self.cutoff:
                continue
            if idx[1] < self.mtx[i, j] and i != j:
                self.mtx[i, j] = idx[1]
                self.mtx[j, i] = idx[1]
                self._registerResiduePair(i, j)

    def _registerResiduePair(self, i, j):
        (resID_A, resID_B) = (self._resArray[i], self._resArray[j]) if i < j else (self._resArray[j], self._resArray[i])
        if resID_A not in self._residuePairRegistry:
            self._residuePairRegistry[resID_A] = [resID_B]
        elif resID_B not in self._residuePairRegistry[resID_A]:
            self._residuePairRegistry[resID_A].append(resID_B)


class ContactOrder(object):
    def __init__(self, struc_name, struc, cutoff=5.0):
        self._contiguous = None
        self.list_CO = []

        for curr_ch in struc.chainList:

            self.sub_struct = struc.chain(curr_ch)
            cm_intra = ContactMap_intra_grid(self.sub_struct,cutoff)
            counter,SUM = 0,0

            for i in range(len(cm_intra._resArray)):
                for j in range(i+1,len(cm_intra._resArray)):
                    if cm_intra.mtx[i, j] < cutoff:
                        SUM += abs(i-j)
                        counter += 1

            self.list_CO.append([struc_name,curr_ch,'{:.2f}'.format(float(SUM)/counter),self.contiguous])

    def __iter__(self):
        for item in self.list_CO:
            yield item

    @property
    def contiguous(self):
        if not self._contiguous:
            self._contiguous = True
            trace_list = [atom for atom in self.sub_struct.trace]
            for i in range(1,len(trace_list)):
                curr_atom = trace_list[i]
                prev_atom = trace_list[i-1]
                dist = minDist([curr_atom],[prev_atom])
                if float(dist) > 4.0:
                    self._contiguous = False
                    break
        return self._contiguous


class interfaceBoolList(object):
    def __init__(self, l1, l2):
        self.l1 = l1
        self.l2 = l2

    def __getitem__(self, k):
        if(k == 0):
            l = self.l1
        elif(k == 1):
            l = self.l2
        else :
            raise ValueError("index \"" + str(k) + "\"out of bounds ")
        s = self.asString(l)
        return s

    def toList(self, k):
        data = []

        if(k == 0):
            l = self.l1
        elif(k == 1):
            l = self.l2
        else :
            raise ValueError("index \"" + str(k) + "\"out of bounds ")
        for d in l:
            data.append({ 'res' : d[0], 'cc' : d[1] })
        return data

    def asString(self, l):
        return '\n'.join([ "%4s %s %s %d" % (d[0]['num'], d[0]['name'], d[0]['chain'], d[1]) for d in l ])



import math,numpy,subprocess

import xml.etree.ElementTree


def kabsch_rmsd(P, Q):
    """
    Rotate matrix P unto Q and calculate the RMSD
    """
    P = kabsch_rotate(P, Q)
    return rmsd(P, Q)


def kabsch_rotate(P, Q):
    """
    Rotate matrix P unto matrix Q using Kabsch algorithm
    """
    U = kabsch(P, Q)

    # Rotate P
    P = numpy.dot(P, U)
    return P


def kabsch(P, Q):
    """
    The optimal rotation matrix U is calculated and then used to rotate matrix
    P unto matrix Q so the minimum root-mean-square deviation (RMSD) can be
    calculated.

    Using the Kabsch algorithm with two sets of paired point P and Q,
    centered around the center-of-mass.
    Each vector set is represented as an NxD matrix, where D is the
    the dimension of the space.

    The algorithm works in three steps:
    - a translation of P and Q
    - the computation of a covariance matrix C
    - computation of the optimal rotation matrix U

    http://en.wikipedia.org/wiki/Kabsch_algorithm

    Parameters:
    P -- (N, number of points)x(D, dimension) matrix
    Q -- (N, number of points)x(D, dimension) matrix

    Returns:
    U -- Rotation matrix

    """

    # Computation of the covariance matrix
    C = numpy.dot(numpy.transpose(P), Q)

    # Computation of the optimal rotation matrix
    # This can be done using singular value decomposition (SVD)
    # Getting the sign of the det(V)*(W) to decide
    # whether we need to correct our rotation matrix to ensure a
    # right-handed coordinate system.
    # And finally calculating the optimal rotation matrix U
    # see http://en.wikipedia.org/wiki/Kabsch_algorithm
    V, S, W = numpy.linalg.svd(C)
    d = (numpy.linalg.det(V) * numpy.linalg.det(W)) < 0.0

    if d:
        S[-1] = -S[-1]
        V[:, -1] = -V[:, -1]

    # Create Rotation matrix U
    U = numpy.dot(V, W)

    return U


def quaternion_rmsd(P, Q):
    """
    based on doi:10.1016/1049-9660(91)90036-O
    Rotate matrix P unto Q and calculate the RMSD
    """
    rot = quaternion_rotate(P, Q)
    P = numpy.dot(P,rot)
    return rmsd(P, Q)


def quaternion_transform(r):
    """
    Get optimal rotation
    note: translation will be zero when the centroids of each molecule are the
    same
    """
    Wt_r = makeW(*r).T
    Q_r = makeQ(*r)
    rot = Wt_r.dot(Q_r)[:3,:3]
    return rot


def makeW(r1,r2,r3,r4=0):
    """
    matrix involved in quaternion rotation
    """
    W = numpy.asarray([
             [r4, r3, -r2, r1],
             [-r3, r4, r1, r2],
             [r2, -r1, r4, r3],
             [-r1, -r2, -r3, r4] ])
    return W


def makeQ(r1,r2,r3,r4=0):
    """
    matrix involved in quaternion rotation
    """
    Q = numpy.asarray([
             [r4, -r3, r2, r1],
             [r3, r4, -r1, r2],
             [-r2, r1, r4, r3],
             [-r1, -r2, -r3, r4] ])
    return Q


def quaternion_rotate(X, Y):
    """
    Calculate the rotation
    """
    N = X.shape[0]
    W = numpy.asarray([makeW(*Y[k]) for k in range(N)])
    Q = numpy.asarray([makeQ(*X[k]) for k in range(N)])
    Qt_dot_W = numpy.asarray([numpy.dot(Q[k].T,W[k]) for k in range(N)])
    W_minus_Q = numpy.asarray([W[k] - Q[k] for k in range(N)])
    C1 = -numpy.sum(Qt_dot_W,axis=0)
    C2 = 0.5*N
    C3 = numpy.sum(W_minus_Q,axis=0)
    A = numpy.dot(C3.T,C3)*C2-C1
    eigen = numpy.linalg.eigh(A)
    r = eigen[1][:,eigen[0].argmax()]
    rot = quaternion_transform(r)
    return rot


def centroid(X):
    """
    Calculate the centroid from a vectorset X
    """

    C = sum(X)/len(X)
    return C


def rmsd(V, W):
    """
    Calculate Root-mean-square deviation from two sets of vectors V and W.
    """
    D = len(V[0])
    N = len(V)
    rmsd = 0.0
    for v, w in zip(V, W):
        rmsd += sum([(v[i]-w[i])**2.0 for i in range(D)])
    return numpy.sqrt(rmsd/N)


def aliFit(structA, structB, aliArrayOne, aliArrayTwo):
    equiResNum = []
    i = -1
    j = -1

    for x,y in zip(aliArrayOne, aliArrayTwo):
        xBool = False
        yBool = False
        if x != '-' :
            xBool = True
            i += 1
        if y != '-' :
            yBool = True
            j += 1

        if not xBool or not yBool:
            continue
        equiResNum.append( { 'iRes' : { 'num' : i , 'name' : x }, 'jRes' : { 'num' : j , 'name' : y } } )

    print("Superimposing on", str(len(equiResNum)), 'CA coordinates')

    P = [ numpy.array( structA.trace[ d['iRes']['num']].coordinates) for d in equiResNum ]
    Q = [ numpy.array( structB.trace[ d['jRes']['num']].coordinates) for d in equiResNum ]

    normal_rmsd = rmsd(P, Q)

    Pc = centroid(P)
    Qc = centroid(Q)
    P -= Pc
    Q -= Qc

    U = kabsch(P, Q)
    #p_all -= Pc
    #p_all = numpy.dot(p_all, U)
    #write_coordinates(p_atoms, p_all, title="{} translated".format(args.structure_a))
    k_rmsd = kabsch_rmsd(P, Q)
    q_rmsd = quaternion_rmsd(P, Q)

    print ("Normal RMSD:", normal_rmsd)
    print ("Kabsch RMSD:", k_rmsd)
    print ("Quater RMSD:", q_rmsd)
    return (U, normal_rmsd, k_rmsd, q_rmsd)

def fit(structObj1, structObj2, **kwargs):

    mode = "blast"
    aliSeq1 = None
    aliSeq2 = None
    if 'mode' in kwargs:
        mode = kwargs['mode']

    if mode == "needle":
        ali = needle(structObj1, structObj2)
        (aliSeq1, aliSeq2) = ali.aaWords

    elif mode == "blast":
        blastThem(structObj1, structObj2, mode)
        msaObj = clustThem(mode)

        aliSeq1 = msaObj[0]['sequence']
        aliSeq2 = msaObj[1]['sequence']

    print (aliSeq1, "\n", aliSeq2)

    _tuple = aliFit(structObj1, structObj2, aliSeq1, aliSeq2)
    return _tuple





import re
import copy
import numpy
import math
import os

aaCode = {
	'ALA' : 'A',
	'ARG' : 'R',
	'ASN' : 'N',
	'ASP' : 'D',
	'GLU' : 'E',
	'GLN' : 'Q',
	'GLY' : 'G',
	'HIS' : 'H',
	'CYS' : 'C',
	'LYS' : 'K',
	'MET' : 'M',
	'LEU' : 'L',
	'ILE' : 'I',
	'MET' : 'M',
	'PRO' : 'P',
	'SER' : 'S',
	'THR' : 'T',
	'TRP' : 'W',
	'TRY' : 'W',
	'TYR' : 'Y',
	'PHE' : 'F',
	'VAL' : 'V',
	'MSE' : 'M'
}

def translate(x):
	global aaCode

	for k,v in aaCode.items():
		if x == v:
			return k
		elif  x == k:
			return v

	if len(x) == 1:
		return 'UNK'
	return 'X'

'''
GL 09/03/2017
 This is minimal implementation of PDB parsing and coordinates mangament

'''

''' PDB format reminder

COLUMNS		DATA  TYPE	FIELD		DEFINITION
-------------------------------------------------------------------------------------
 1 -  6		Record name   "ATOM  "
 7 - 11		Integer	   serial	   Atom  serial number.
13 - 16		Atom		  name		 Atom name.
17			 Character	 altLoc	   Alternate location indicator.
18 - 20		Residue name  resName	  Residue name.
22			 Character	 chainID	  Chain identifier.
23 - 26		Integer	   resSeq	   Residue sequence number.
27			 AChar		 iCode		Code for insertion of residues.
31 - 38		Real(8.3)	 x			Orthogonal coordinates for X in Angstroms.
39 - 46		Real(8.3)	 y			Orthogonal coordinates for Y in Angstroms.
47 - 54		Real(8.3)	 z			Orthogonal coordinates for Z in Angstroms.
55 - 60		Real(6.2)	 occupancy	Occupancy.
61 - 66		Real(6.2)	 tempFactor   Temperature  factor.
77 - 78		LString(2)	element	  Element symbol, right-justified.
79 - 80		LString(2)	charge	   Charge  on the atom.

'''

class Parser(object):
	def __init__(self):
		pass

	def load(self, heteroatoms = False, **kwargs):
		structureObj = Structure();

		def _read(stream):
			currentAtomList = structureObj.pop();
			newModel = False
			for l in stream:
				#print ">>" + l + "<<"
				if l.startswith("ATOM ") or (heteroatoms and l.startswith("HETATM")):
					if newModel:
						currentAtomList = structureObj.pop()
						newModel = False
					currentAtomList.append(Atom(string=l))
				elif l.startswith("ENDMDL"):
					newModel = True

				if l.startswith("SEQRES "):
					structureObj.loadSEQRES(l)

		if 'file' in kwargs:
			with open(kwargs['file'], 'r') as f:
				_read(f)

		if 'stream' in kwargs:
			#print kwargs['stream'] + "<--"
			_read(kwargs['stream'].splitlines(True))

		if len(structureObj.model[0]) == 0:
			raise ValueError("Empty coordinates")

		# Assign a name
		if not structureObj.name:
			if 'file' in kwargs:
				base = os.path.basename(kwargs['file'])
				structureObj.name = os.path.splitext(base)[0]

		if not structureObj.name:
			structureObj.name = "a protein"

		return structureObj



class Model(object):
	def init(self):
		pass



class Structure(object):
	def __init__(self):
		self.name = None
		self.model = []
		self._trace = None
		self._fasta = None
		self.currModel = 1
		self.bufferSelection = []
		self._chainList = None
		self._residues = None
		self._resID = None
		self.SEQRES = {}

	def __len__(self):
		return len(self.model[self.currModel - 1])

	def loadSEQRES(self, line):

		aaSeq = line.split()[2:]
		segID = aaSeq.pop(0)
		aaSeq.pop(0)
		if segID not in self.SEQRES:
			self.SEQRES[segID] = ''
		
		self.SEQRES[segID] += ''.join([ translate(aa) for aa in aaSeq ])
		


	def setCoordinateFromDictorize(self, dictorizedSelf):
		atomList = self.model[self.currModel - 1]
		#print atomList
		#print len(atomList)
		#print len(dictorizedSelf['x'])
		for i in range(len(atomList)):
	#        print i
			self.model[self.currModel - 1][i].x = dictorizedSelf['x'][i]
			self.model[self.currModel - 1][i].y = dictorizedSelf['y'][i]
			self.model[self.currModel - 1][i].z = dictorizedSelf['z'][i]

	@property
	def atomDictorize(self):
		atomList = self.model[self.currModel - 1]
		(x, y, z, seqRes, chainID, resName, name) = self.atomVectorize;
		return { "x" : x,
				 "y" : y,
				 "z" : z,
				 "seqRes" : seqRes,
				 "chainID" : chainID,
				 "resName" : resName,
				 "name" : name
				}

	@property
	def atomVectorize(self):
		atomList = self.model[self.currModel - 1]
		seqRes = []
		chainID = []
		x = []
		y = []
		z = []
		resName = []
		name = []
		for a in atomList:
			seqRes.append(a.seqRes)
			chainID.append(a.chainID)
			x.append(a.x)
			y.append(a.y)
			z.append(a.z)
			name.append(a.name)
			resName.append(a.resName)
		return (x, y, z, seqRes, chainID, resName, name)

	@property
	def getResID(self):
		if not self._resID:
			self._resID = [residue.data[0].getResID for residue in self.byres()]
		return self._resID
	# Browse model and create a new pdbObject with a sinlge model that satifise the provided conditions
	# Eg: ask for the 1st model w/ oligomeric composition chain=["A","B","C"]
	def modelReduce(self, **kwargs):
		modelMatch = -1

		if 'chain' in kwargs:
			chainList = kwargs['chain']
			for i, model  in enumerate(self.model):
				for segID in chainList:
					found = False
					for p_atom in model:
						if p_atom.chainID == segID:
							print(p_atom)
							found = True
							break
					if not found:
						break
				if found:
					modelMatch = i
					break
		if modelMatch < 0:
			return None

		print(" Model ", str(modelMatch), "satisfies conditions")

		cloneObj = Structure()
		clonedModel =  self.model[modelMatch]
		currentAtomList = cloneObj.pop()
		for p_atom in clonedModel:
			currentAtomList.append(copy.deepcopy(p_atom))
		return cloneObj

	def peptideSeed(self): # A function to operate w/ pyprotein utilities like nw_custom

		return {
			'id' : self.name,
			'desc' : 'pdb file fasta translation',
			'seq' : self.fasta
		}

# move a structure along a vector
	def nudge(self, C):
		V = [ numpy.array( d.coordinates ) for d in self.atomRecord ]
		V += C
		for i, a in enumerate (V):
			self.model[self.currModel - 1][i].x = V[i,0]
			self.model[self.currModel - 1][i].y = V[i,1]
			self.model[self.currModel - 1][i].z = V[i,2]


	def centerOrigin(self):
		V = [ numpy.array( d.coordinates ) for d in self.atomRecord ]
	# move to centroid
		C = sum(V)/len(V)
		print(C)
		V -= C
		for i, a in enumerate (V):
			#print
			self.model[self.currModel - 1][i].x = V[i,0]
			self.model[self.currModel - 1][i].y = V[i,1]
			self.model[self.currModel - 1][i].z = V[i,2]

		return C



	def rotate(self, **kwargs):
# 1st we move centroid to origin
# 2nd we rotate
# 3rd we move centroid back to original
# U=RotationMatrix  OR (alpha=0, beta=0, gamma=0)
		centeringBool = True

		if ('nocenter' in kwargs):
			if kwargs['nocenter']:
				centeringBool = False

		if 'U' not in kwargs and ('alpha' not in kwargs or 'beta' not in kwargs or 'gamma' not in kwargs):
			raise ValueError("Specify a 3x3 rotation matrix or alpha,beta,angle for rotation around X,Y and Z axis");

		V = [ numpy.array( d.coordinates ) for d in self.atomRecord ]

		if centeringBool:
	# move to centroid
			C = sum(V)/len(V)
			V -= C

		if 'U' not in kwargs:
			alpha=kwargs['alpha']
			beta=kwargs['beta']
			gamma=kwargs['gamma']

	# Rotate around origin
			Rx = numpy.matrix([
						[1,          0,           0],
						[0, math.cos(alpha), -math.sin(alpha)],
						[0, math.sin(alpha),  math.cos(alpha)]
					   ])
			Ry = numpy.matrix([
						[math.cos(beta),  0,  math.sin(beta)],
						[0,          1,          0],
						[-math.sin(beta), 0,  math.cos(beta)]
						])
			Rz = numpy.matrix([
						[math.cos(gamma),  -math.sin(gamma), 0],
						[math.sin(gamma),  math.cos(gamma), 0],
						[0,                    0, 1]
						])

		#print V[0]


		for i, a in enumerate (V):

			v = numpy.matrix([[a[0]],
						   [a[1]],
						   [a[2]]])
			if 'U' not in kwargs:
				v_ = Rz*Ry*Rx*v
			else :
				v_ = kwargs['U']*v

			self.model[self.currModel - 1][i].x = v_[0,0]
			self.model[self.currModel - 1][i].y = v_[1,0]
			self.model[self.currModel - 1][i].z = v_[2,0]


	## Move back
	 #   t = numpy.matrix([[C[0]],
					   #[C[1]],
					   #[C[2]]])
	  #  tBack = t
		if centeringBool:
			for a in self.model[self.currModel - 1]:
				a.x += C[0]
				a.y += C[1]
				a.z += C[2]
		#print self.model[self.currModel - 1][0]


	def __str__(self):
		asString = '' # some header here
		for n, model in enumerate(self.model):
			self.currModel = n + 1
			asString += ''.join([ str(a) if str(a).endswith("\n") else str(a) + '\n' for a in self.atomRecord ])
			#asString += "ENDMDL\n"
		return asString

	def clone(self):
		cloneObj = Structure()

		for struct in self.model:
			currentAtomList = cloneObj.pop();
			for p_atom in struct:
				currentAtomList.append(copy.deepcopy(p_atom))

		return cloneObj

	@property
	def trace(self):
		if not self._trace:
			self._trace = []
			for r in self.byres():
				for a in r:
					if a.name == 'CA':
						self._trace.append(a)

		return self._trace

	@property
	def residueNumber(self):
		return len([ res for res in self.byres() ])

	# Return consecutive 1letter amino-acid for residues with defined Calpha coordinates
	@property
	def fasta(self):
		if not self._fasta:
			self._fasta = ''.join([ r.fasta for r in self.byres(strict=True) ])
		return self._fasta

	@property
	def atomRecord(self):
		for atom in self.model[self.currModel - 1]:
			yield atom

	def atomFromResname(self, resname):
		for atom in self.model[self.currModel - 1]:
			if atom.resName == resname:
				yield atom

	@property
	def chainList(self):
		if not self._chainList:
			buf = [ r.chain for r in self.byres() ]
			self._chainList = list(set(buf))
		return self._chainList

	def byres(self, strict=False):

		i = 0
		data = self.model[self.currModel - 1]
		w_resSeq = data[0].resSeq
		w_chainID = data[0].chainID
		w_iCode = data[0].iCode
		i_start = 0
		i_stop = 0
		while i < len(data):
			#print str(i) + " -- " + str(len(data))
			cur_atom = data[i]
			#print str(w_resSeq) + ' vs ' + str(cur_atom.resSeq) + ' and ' + w_chainID + ' vs ' + cur_atom.chainID
			if w_resSeq != cur_atom.resSeq or w_chainID != cur_atom.chainID or w_iCode != cur_atom.iCode:
				#i_stop = i + 1
				#print "pulling at " + str(i_start) + ":" + str(i)
				x = Residue(data[i_start:i])
				if (strict and x.hasCalpha) or not strict:
					yield x
				w_resSeq  = cur_atom.resSeq
				w_chainID = cur_atom.chainID
				w_iCode = cur_atom.iCode

				i_start = i
			i += 1

		x = Residue(data[i_start:i])
		if (strict and x.hasCalpha) or not strict:
			yield x

	def pop(self): # Append an empty array to model list and return its referenc

		self.model.append([])
		return self.model[-1]

	# Return a pdb object limited to specified chain ID
	def chain(self, chain):
		chainList = []
		try:
			chainList = [e for e in chain]
		except TypeError:
			chainList.append(chain)

		cloneObj = Structure()

		for struct in self.model:
			currentAtomList = cloneObj.pop();
			for p_atom in struct:
				if p_atom.chainID in chainList:
					currentAtomList.append(copy.deepcopy(p_atom))

		cloneObj.name = self.name + ':' + ','.join(chainList)
		if len(cloneObj.model[0]) == 0:
			print("Empty selection for chain(s) \"", str(chain), "\"")
			return None

		return cloneObj


	def select(exp): # TO DO
		pass
		#
		# Should implment binary tree construction from the parsing of pymol-like expression
		# All tree relationship could then be mapped to set operation on atoms lists
		#
		#        http://pyparsing.wikispaces.com/

class Residue(object):
	def __init__(self, atomArray):
		self.data = atomArray

	def __len__(self):
		return len(self.data)

	def __repr__(self):
	   return '\n'.join([ str(a) for a in self.data ])

	def __getitem__(self, key):
		return self.data[key]

	def __iter__(self):
		for d in self.data:
			yield d

	def __str__(self):
		return "%3s %4s%c %s" %(self.name, self.num.strip(), self.iCode, self.chain)

	def __eq__(self, other):
		return (self.chain, self.seqRes) == (other.chain, other.seqRes)

	def __ne__(self, other):
		return not(self == other)

	@property
	def fasta(self):
		return translate(self.data[0].resName)

	@property
	def id(self):
		return self.data[0].resName + str(self.data[0].resSeq) + ':' + self.data[0].chainID

	@property
	def name(self):
		return self.data[0].resName

	@property
	def seqRes(self):
		return self.data[0].seqRes

	@property
	def num(self):
		return self.data[0].resSeq.strip()

	@property
	def chain(self):
		return self.data[0].chainID

	@property
	def iCode(self):
		return self.data[0].iCode

	@property
	def seqRes(self):
		return self.data[0].seqRes

	@property
	def hasCalpha(self):
		for a in self:
			if a.name == "CA":
				return True
		return False

	def asPdbRecord(self):
		return self.__repr__()

class Atom(object):
	def __init__(self, **kwargs):
		if 'string' in kwargs:
			buf=kwargs['string'].rstrip()
			bufLen = len(buf)
			if bufLen < 54:
				raise TypeError("Unexpected atom record length \"" + str(bufLen) + "\" on input at \n" + buf)
			self.recordName=buf[0:6].replace(" ", "")
			self.serial=int(buf[6:11])
			self.name=buf[12:16].replace(" ", "")
			self.altLoc= buf[16]
			self.resName=buf[17:20].replace(" ", "")
			self.chainID=buf[21]
			self.resSeq=buf[22:26] # Not sure
			self.iCode=buf[26]
			self.x=float(buf[30:38])
			self.y=float(buf[38:46])
			self.z=float(buf[46:54])
			self.occupancy  = float(buf[54:60])           if bufLen > 54 else None
			self.tempFactor = float(buf[60:66])           if bufLen > 60 else None
			self.element    = buf[76:78].replace(" ", "") if bufLen > 76 else None
			self.charge     = buf[78:80].replace(" ", "") if bufLen > 78 else None

	@property
	def seqRes(self):
		return str(self.resSeq.lstrip()) + str(self.iCode)

	@property
	def coordinates(self):
		return [self.x, self.y, self.z]

	@property
	def toVector(self):
		return [self.x, self.y, self.z, self.resSeq + self.iCode, self.chainID]

	@property
	def getResID(self):
	#   return self.resName + str(self.resSeq) + ':' + self.chainID + ':' + self.iCode
		return self.resName + " " + self.seqRes + ':' + self.chainID

	def __hash__(self):
		tup = (self.recordName, self.serial, self.altLoc, self.resName, self.chainID, self.resSeq, self.iCode, self.x, self.y, self.z, self.occupancy, self.tempFactor, self.element, self.charge)
		return hash(tup)

	def __str__(self):
		if self.charge:
			return "%-5s%5d %4s%s%4s %s%4s%s   %8.3f%8.3f%8.3f%6.2f%6.2f          %2s%2s" % (self.recordName, self.serial, self.name, self.altLoc, self.resName, self.chainID, self.resSeq,self.iCode, self.x, self.y, self.z, self.occupancy, self.tempFactor, self.element, self.charge)
		if self.element:
			return "%-5s%5d %4s%s%4s %s%4s%s   %8.3f%8.3f%8.3f%6.2f%6.2f          %2s  " % (self.recordName, self.serial, self.name, self.altLoc, self.resName, self.chainID, self.resSeq,self.iCode, self.x, self.y, self.z, self.occupancy, self.tempFactor, self.element)
		if self.tempFactor:
			return "%-5s%5d %4s%s%4s %s%4s%s   %8.3f%8.3f%8.3f%6.2f%6.2f              " % (self.recordName, self.serial, self.name, self.altLoc, self.resName, self.chainID, self.resSeq,self.iCode, self.x, self.y, self.z, self.occupancy, self.tempFactor)
		if self.occupancy:
			return "%-5s%5d %4s%s%4s %s%4s%s   %8.3f%8.3f%8.3f%6.2f                 " % (self.recordName, self.serial, self.name, self.altLoc, self.resName, self.chainID, self.resSeq,self.iCode, self.x, self.y, self.z, self.occupancy)
		return "%-5s%5d %4s%s%4s %s%4s%s   %8.3f%8.3f%8.3f                          " % (self.recordName, self.serial, self.name, self.altLoc, self.resName, self.chainID, self.resSeq,self.iCode, self.x, self.y, self.z)
